@@ -10,22 +10,50 @@ BEGIN {
 	@ISA = qw(Exporter);
 }
 
-$VERSION = "0.02";
+$VERSION = "0.03";
 
 $VERSION = eval $VERSION;
 
-@EXPORT = @EXPORT_OK = qw(try catch);
+@EXPORT = @EXPORT_OK = qw(try catch finally);
 
-sub try (&;$) {
-	my ( $try, $catch ) = @_;
+$Carp::Internal{+__PACKAGE__}++;
+
+# Need to prototype as @ not $$ because of the way Perl evaluates the prototype.
+# Keeping it at $$ means you only ever get 1 sub because we need to eval in a list
+# context & not a scalar one
+
+sub try (&;@) {
+	my ( $try, @code_refs ) = @_;
 
 	# we need to save this here, the eval block will be in scalar context due
 	# to $failed
 	my $wantarray = wantarray;
 
+	my ( $catch, $finally );
+
+	# find labeled blocks in the argument list.
+	# catch and finally tag the blocks by blessing a scalar reference to them.
+	foreach my $code_ref (@code_refs) {
+		next unless $code_ref;
+
+		my $ref = ref($code_ref);
+
+		if ( $ref eq 'Try::Tiny::Catch' ) {
+			$catch = ${$code_ref};
+		} elsif ( $ref eq 'Try::Tiny::Finally' ) {
+			$finally = ${$code_ref};
+		} else {
+			use Carp;
+			confess("Unknown code ref type given '${ref}'. Check your usage & try again");
+		}
+	}
+
+	# save the value of $@ so we can set $@ back to it in the beginning of the eval
+	my $prev_error = $@;
+
 	my ( @ret, $error, $failed );
 
-	# FIXME consider using local $SIG{__DIE__} to accumilate all errors. It's
+	# FIXME consider using local $SIG{__DIE__} to accumulate all errors. It's
 	# not perfect, but we could provide a list of additional errors for
 	# $catch->();
 
@@ -37,6 +65,7 @@ sub try (&;$) {
 		# failed will be true if the eval dies, because 1 will not be returned
 		# from the eval body
 		$failed = not eval {
+			$@ = $prev_error;
 
 			# evaluate the try block in the correct context
 			if ( $wantarray ) {
@@ -50,20 +79,26 @@ sub try (&;$) {
 			return 1; # properly set $fail to false
 		};
 
-		# copy $@ to $error, when we leave this scope local $@ will revert $@
+		# copy $@ to $error; when we leave this scope, local $@ will revert $@
 		# back to its previous value
 		$error = $@;
 	}
 
-	# at this point $failed contains a true value if the eval died even if some
-	# destructor overwrite $@ as the eval was unwinding.
+	# at this point $failed contains a true value if the eval died, even if some
+	# destructor overwrote $@ as the eval was unwinding.
 	if ( $failed ) {
 		# if we got an error, invoke the catch block.
 		if ( $catch ) {
 			# This works like given($error), but is backwards compatible and
 			# sets $_ in the dynamic scope for the body of C<$catch>
 			for ($error) {
-				return $catch->($error);
+				my $catch_return = $catch->($error);
+
+				# Finally blocks run after all other blocks so it is executed here
+				$finally->() if ( $finally );
+
+				#And return whatever catch returned
+				return $catch_return;
 			}
 
 			# in case when() was used without an explicit return, the C<for>
@@ -72,15 +107,31 @@ sub try (&;$) {
 
 		return;
 	} else {
+		# Execute finally block once we decided we worked
+		$finally->() if ( $finally );
+
 		# no failure, $@ is back to what it was, everything is fine
 		return $wantarray ? @ret : $ret[0];
 	}
 }
 
-sub catch (&) {
-	return $_[0];
+sub catch (&;@) {
+	my ( $block, @rest ) = @_;
+
+	return (
+		bless(\$block, 'Try::Tiny::Catch'),
+		@rest,
+	);
 }
 
+sub finally (&;@) {
+	my ( $block, @rest ) = @_;
+
+	return (
+		bless(\$block, 'Try::Tiny::Finally'),
+		@rest,
+	);
+}
 
 __PACKAGE__
 
@@ -108,7 +159,7 @@ Try::Tiny - minimal try/catch with proper localization of $@
 
 =head1 DESCRIPTION
 
-This module provides bare bones C<try>/C<catch> statements that are designed to
+This module provides bare bones C<try>/C<catch>/C<finally> statements that are designed to
 minimize common mistakes with eval blocks, and NOTHING else.
 
 This is unlike L<TryCatch> which provides a nice syntax and avoids adding
@@ -126,18 +177,37 @@ It's designed to work as correctly as possible in light of the various
 pathological edge cases (see L<BACKGROUND>) and to be compatible with any style
 of error values (simple strings, references, objects, overloaded objects, etc).
 
+If the try block dies, it returns the value of the last statement executed in
+the catch block, if there is one. Otherwise, it returns C<undef> in scalar
+context or the empty list in list context. The following two examples both
+assign C<"bar"> to C<$x>.
+
+	my $x = try { die "foo" } catch { "bar" };
+
+	my $x = eval { die "foo" } || "bar";
+
+You can add finally blocks making the following true.
+
+	my $x;
+	try { die 'foo' } finally { $x = 'bar' };
+	try { die 'foo' } catch { warn "Got a die: $_" } finally { $x = 'bar' };
+
+Finally blocks are always executed making them suitable for cleanup code
+which cannot be handled using local.
+
 =head1 EXPORTS
 
 All functions are exported by default using L<Exporter>.
 
-In the future L<Sub::Exporter> may be used to allow the keywords to be renamed,
-but this technically does not satisfy Adam Kennedy's definition of "Tiny".
+If you need to rename the C<try>, C<catch> or C<finally> keyword consider using
+L<Sub::Import> to get L<Sub::Exporter>'s flexibility.
 
 =over 4
 
-=item try (&;$)
+=item try (&;@)
 
-Takes one mandatory try subroutine and one optional catch subroutine.
+Takes one mandatory try subroutine, an optional catch subroutine & finally
+subroutine.
 
 The mandatory subroutine is evaluated in the context of an C<eval> block.
 
@@ -149,19 +219,52 @@ with the error in C<$_> (localized) and as that block's first and only
 argument.
 
 Note that the error may be false, but if that happens the C<catch> block will
-still be invoked..
+still be invoked.
 
-=item catch (&)
+Once all execution is finished then the finally block if given will execute.
+
+=item catch (&;$)
 
 Intended to be used in the second argument position of C<try>.
 
-Just returns the subroutine it was given.
+Returns a reference to the subroutine it was given but blessed as
+C<Try::Tiny::Catch> which allows try to decode correctly what to do
+with this code reference.
 
 	catch { ... }
 
-is the same as
+Inside the catch block the previous value of C<$@> is still available for use.
+This value may or may not be meaningful depending on what happened before the
+C<try>, but it might be a good idea to preserve it in an error stack.
 
-	sub { ... }
+=item finally (&;$)
+
+  try     { ... }
+  catch   { ... }
+  finally { ... };
+
+Or
+
+  try     { ... }
+  finally { ... };
+
+Or even
+
+  try     { ... }
+  finally { ... }
+  catch   { ... };
+
+Intended to be the second or third element of C<try>. Finally blocks are always
+executed in the event of a successful C<try> or if C<catch> is run. This allows
+you to locate cleanup code which cannot be done via C<local()> e.g. closing a file
+handle.
+
+B<You must always do your own error handling in the finally block>. C<Try::Tiny> will
+not do anything about handling possible errors coming from code located in these
+blocks.
+
+In the same way C<catch()> blesses the code reference this subroutine does the same
+except it bless them as C<Try::Tiny::Finally>.
 
 =back
 
@@ -180,12 +283,19 @@ not yet handled.
 C<$@> must be properly localized before invoking C<eval> in order to avoid this
 issue.
 
+More specifically, C<$@> is clobbered at the begining of the C<eval>, which
+also makes it impossible to capture the previous error before you die (for
+instance when making exception objects with error stacks).
+
+For this reason C<try> will actually set C<$@> to its previous value (before
+the localization) in the beginning of the C<eval> block.
+
 =head2 Localizing $@ silently masks errors
 
 Inside an eval block C<die> behaves sort of like:
 
 	sub die {
-		$@_ = $_[0];
+		$@ = $_[0];
 		return_undef_from_eval();
 	}
 
@@ -213,8 +323,8 @@ This code is wrong:
 
 because due to the previous caveats it may have been unset.
 
-C<$@> could also an overloaded error object that evaluates to false, but that's
-asking for trouble anyway.
+C<$@> could also be an overloaded error object that evaluates to false, but
+that's asking for trouble anyway.
 
 The classic failure mode is:
 
@@ -233,7 +343,7 @@ The classic failure mode is:
 	}
 
 In this case since C<Object::DESTROY> is not localizing C<$@> but still uses
-C<eval> it will set C<$@> to C<"">.
+C<eval>, it will set C<$@> to C<"">.
 
 The destructor is called when the stack is unwound, after C<die> sets C<$@> to
 C<"foo at Foo.pm line 42\n">, so by the time C<if ( $@ )> is evaluated it has
@@ -276,15 +386,53 @@ concisely match errors:
 
 =item *
 
+C<@_> is not available, you need to name your args:
+
+	sub foo {
+		my ( $self, @args ) = @_;
+		try { $self->bar(@args) }
+	}
+
+=item *
+
+C<return> returns from the C<try> block, not from the parent sub (note that
+this is also how C<eval> works, but not how L<TryCatch> works):
+
+	sub bar {
+		try { return "foo" };
+		return "baz";
+	}
+
+	say bar(); # "baz"
+
+=item *
+
 C<try> introduces another caller stack frame. L<Sub::Uplevel> is not used. L<Carp>
 will report this when using full stack traces. This lack of magic is considered
 a feature.
 
 =item *
 
-The value of C<$_> in the C<catch> block is not guaranteed to be preserved,
-there is no safe way to ensure this if C<eval> is used unhygenically in
-destructors. It's only guaranteeed that the C<catch> will be called.
+The value of C<$_> in the C<catch> block is not guaranteed to be the value of
+the exception thrown (C<$@>) in the C<try> block.  There is no safe way to
+ensure this, since C<eval> may be used unhygenically in destructors.  The only
+guarantee is that the C<catch> will be called if an exception is thrown.
+
+=item *
+
+The return value of the C<catch> block is not ignored, so if testing the result
+of the expression for truth on success, be sure to return a false value from
+the C<catch> block:
+
+	my $obj = try {
+		MightFail->new;
+	} catch {
+		...
+
+		return; # avoid returning a true value;
+	};
+
+	return unless $obj;
 
 =back
 
@@ -317,10 +465,21 @@ Provides a C<catch> statement, but properly calling C<eval> is your
 responsibility.
 
 The C<try> keyword pushes C<$@> onto an error stack, avoiding some of the
-issues with C<$@> but you still need to localize to prevent clobbering.
+issues with C<$@>, but you still need to localize to prevent clobbering.
 
 =back
 
+=head1 LIGHTNING TALK
+
+I gave a lightning talk about this module, you can see the slides (Firefox
+only):
+
+L<http://nothingmuch.woobling.org/talks/takahashi.xul?data=try_tiny.txt>
+
+Or read the source:
+
+L<http://nothingmuch.woobling.org/talks/yapc_asia_2009/try_tiny.yml>
+
 =head1 VERSION CONTROL
 
 L<http://github.com/nothingmuch/try-tiny/>
